(ns age-of-sail.core
  (:require [age-of-sail.vec2 :refer :all]
            [seesaw.core :refer :all]
            [seesaw.bind :as b]
            [clojure.math :as math]))

(def tickrate 100)
(def hardcoded-wind [0.1 3.0]) ;; A strong easternly wind!
(defonce ships (atom []))

(defn ship-names
  "Gets the names from a list of ships"
  [ships]
  (map #(:name @%) ships))

(defn find-ship
  "Finds a ship with the given name"
  [ships name]
  (some #(when (= name (:name @%)) %) ships))

(defn downwind-force
  "Calculates the force for a single square downwind sail
  Force = furl × length² × dot(wind, heading)"
  [sail wind heading]
  (scale heading (* (:furl sail) (:length sail) (:length sail) (dot wind heading))))

(defn downwind-sails-force
  [ship wind]
  (loop [slots (:slots ship)
         force (zero)]
    (if-let [slot (first slots)]
      (if (= :downwind-sail (:type slot))
        (recur (rest slots) (add force (downwind-force slot wind (:heading ship))))
        (recur (rest slots) force))
      force)))

(defn physics-step
  [ship wind]
  (dosync
   ;; Update position
   (alter ship update :position add (scale (:velocity @ship) (/ tickrate)))
   ;; linear dampening
   (alter ship update :velocity scale (- 1.0 (/ 0.5 tickrate)))
   ;; wind force
   (alter ship update :velocity add (scale (downwind-sails-force @ship wind) (/ tickrate)))))

(defn tick
  [ships]
  (doseq [ship ships]
    (-> ship (physics-step hardcoded-wind))))

;; Simulation controls
(defonce program (atom :stopped))
(defn game-loop
  []
  (while (#{:running :paused} @program)
    (when (= :running @program)
      (tick @ships))
    (Thread/sleep (quot 1000 tickrate)))
  (when-not (compare-and-set! program :killed :stopped)
    (throw "Error: tried to stop a program that wasn't killed!")))

(defn pause-program
  []
  (compare-and-set! program :running :paused))

(defn kill-program
  []
  (compare-and-set! program :running :paused)
  (compare-and-set! program :paused :killed))

(defn start-program
  []
  (when (= (first (reset-vals! program :running)) :stopped)
    (.start (Thread. game-loop))))

;; UI
(defn ignore-args
  [f]
  (fn [& _] (f)))

(def tracked-ship (atom nil))

(defn simulation-controls
  []
  (let [start-button (button :text "Start" :listen [:mouse-clicked (ignore-args start-program)])
        pause-button (button :text "Pause" :listen [:mouse-clicked (ignore-args pause-program)])
        kill-button (button :text "Kill" :listen [:mouse-clicked (ignore-args kill-program)])]
    (letfn [(start-enabled? [state] (contains? #{:stopped :paused} state))
            (pause-enabled? [state] (= :running state))
            (kill-enabled? [state] (contains? #{:running :paused} state))]
      (b/bind program (b/tee
                       (b/bind (b/transform start-enabled?) (b/property start-button :enabled?))
                       (b/bind (b/transform pause-enabled?) (b/property pause-button :enabled?))
                       (b/bind (b/transform kill-enabled?) (b/property kill-button :enabled?))))
      (config! start-button :enabled? (start-enabled? @program))
      (config! pause-button :enabled? (pause-enabled? @program))
      (config! kill-button :enabled? (kill-enabled? @program))
      (horizontal-panel :items [start-button pause-button kill-button]))))

(defn ship-chooser
  []
  (let [name (combobox :model (ship-names @ships))]
    (b/bind ships (b/transform ship-names) (b/property name :model))
    (b/bind (b/selection name) (b/transform #(find-ship @ships %)) tracked-ship)
    (flow-panel :items ["Ship Name" name])))

(defn format-position
  [ship]
  (apply format "X: %.2f Y: %.2f" (:position @ship)))

(defn ship-info
  []
  (let [panel (vertical-panel :visible? false)
        position (label)]
    (b/bind tracked-ship (b/some identity)
            (b/tee
             (b/bind (b/transform format-position) (b/value position))))
    (add! panel (flow-panel :items ["Position" position]))
    panel))

(defn show-ui
  []
  (let [root (frame :title "Age of Sail" :content (vertical-panel))
        chooser (ship-chooser)
        info (ship-info)]
    (b/bind tracked-ship (b/transform boolean) (b/tee (b/property info :visible?)
                                                      (b/b-do [_] (pack! root))))
    (doto root
      (add! (simulation-controls))
      (add! chooser)
      (add! info)
      pack!
      show!)))

(defn -main
  [args]
  (start-program)
  (show-ui))
